
#include "V3.h"
#include "M33.h"

using namespace std;

#include <ostream>

V3::V3(float x, float y, float z) {

	xyz[0] = x;
	xyz[1] = y;
	xyz[2] = z;

}

float& V3::operator[](int i) {

	return xyz[i];

}

V3 V3::operator+(V3 op1) {

	V3 ret;
	V3 &op0 = (*this);
	ret[0] = op0[0] + op1[0];
	ret[1] = op0[1] + op1[1];
	ret[2] = op0[2] + op1[2];
	return ret;

}

V3 V3::operator-(V3 op1) {

	V3 ret;
	V3 &op0 = (*this);
	ret[0] = op0[0] - op1[0];
	ret[1] = op0[1] - op1[1];
	ret[2] = op0[2] - op1[2];
	return ret;

}

float V3::operator*(V3 op1) {

	V3 &op0 = (*this);
	float ret;
	ret = op0[0] * op1[0] + op0[1] * op1[1] + op0[2] * op1[2];
	return ret;

}

V3 V3::operator/(float scf) {

	V3& v = *this;
	return v * (1.0f / scf);

}

V3 V3::operator*(float scf) {

	V3 ret;
	V3& v = *this;
	ret[0] = v[0] * scf;
	ret[1] = v[1] * scf;
	ret[2] = v[2] * scf;
	return ret;

}


V3 V3::operator^(V3 v2) {
	// y1z2 - z1y2
	// z1x2 - x1z2
	// x1y2 - y1x2

	V3 &v1 = *this;
	V3 ret;
	ret[0] = v1[1] * v2[2] - v1[2] * v2[1];
	ret[1] = v1[2] * v2[0] - v1[0] * v2[2];
	ret[2] = v1[0] * v2[1] - v1[1] * v2[0];
	return ret;

}

ostream& operator<<(ostream& ostr, V3 v) {

	return ostr << v[0] << " " << v[1] << " " << v[2];

}


void V3::SetColor(unsigned int color) {

	unsigned char *rgba = (unsigned char*)&color;
	V3 &v = *this;
	v[0] = (float)rgba[0] / 255.0f;
	v[1] = (float)rgba[1] / 255.0f;
	v[2] = (float)rgba[2] / 255.0f;

}

unsigned int V3::GetColor() {

	unsigned int ret;
	unsigned char *rgba = (unsigned char*)&ret;
	V3 &v = *this;
	for (int ci = 0; ci < 3; ci++) {
		if (v[ci] < 0.0f)
			rgba[ci] = 0;
		else if (v[ci] >= 1.0f)
			rgba[ci] = 255;
		else
			rgba[ci] = (unsigned char)(v[ci] * 255.0f);
	}
	rgba[3] = 0xFF;
	return ret;

}

V3 V3::RotateThisVectorAboutAxis(V3 aD, float theta) {

	return RotateThisPointAboutAxis(V3(0.0f, 0.0f, 0.0f), aD, theta);

}

V3 V3::RotateThisPointAboutAxis(V3 aO, V3 aD, float theta) {

	V3 aux;
	if (fabsf(aD[0]) < fabsf(aD[1])) {
		aux = V3(1.0f, 0.0f, 0.0f);
	}
	else {
		aux = V3(0.0f, 1.0f, 0.0f);
	}

	V3 xa = aD ^ aux;
	V3 ya = aD;
	V3 za = xa ^ ya;

	M33 lcs; lcs[0] = xa; lcs[1] = ya; lcs[2] = za;

	V3 &P = *this;
	V3 lP = lcs*(P - aO);
	M33 rotY; rotY.SetRotY(theta);
	V3 rlP = rotY*lP;
	V3 ret = lcs.Inverted()*rlP + aO;
	return ret;
}


float V3::Length() {

	V3 &v = *this;
	return sqrtf(v*v);
}


V3 V3::UnitVector() {

	V3 &v = *this;
	return v/v.Length();

}

// n and l have to be unit vectors
V3 V3::Light(V3 n, V3 l, float ka) {

	V3 ret;
	V3 &v = *this;

	float kd = l*n;
	kd = (kd > 0.0f) ? kd : 0.0f;
	ret = v*(ka + (1.0f - ka)*kd);

	return ret;

}

V3 V3::Light(V3 P, V3 n, V3 L, float ka) {

	return Light(n, (L - P).UnitVector(), ka);

}

